@@ -3,7 +3,6 @@ from cfelpyutils.crystfel_utils import load_crystfel_geometry
 import h5py
 from itertools import chain, product
 import numpy as np
-from scipy.ndimage import affine_transform
 import warnings
 
 from .crystfel_fmt import write_crystfel_geom
@@ -33,7 +32,7 @@ class GeometryFragment:
     @classmethod
     def from_panel_dict(cls, d):
         res = d['res']
-        corner_pos = np.array([d['cnx'], d['cny'], d['coffset']]) / res
+        corner_pos = np.array([d['cnx']/res, d['cny']/res, d['coffset']])
         ss_vec = np.array([d['ssx'], d['ssy'], d['ssz']]) / res
         fs_vec = np.array([d['fsx'], d['fsy'], d['fsz']]) / res
         ss_pixels = d['max_ss'] - d['min_ss'] + 1
@@ -82,6 +81,7 @@ class DetectorGeometryBase:
     pixel_size = 0.0
     frag_ss_pixels = 0
     frag_fs_pixels = 0
+    n_quads = 0
     n_modules = 0
     n_tiles_per_module = 0
     expected_data_shape = (0, 0, 0)
@@ -233,6 +233,32 @@ class DetectorGeometryBase:
                 transform=ax.transAxes)
         return ax
 
+    @classmethod
+    def _cfel_panels_by_data_coord(cls, panels: dict):
+        """Arrange panel dicts from CrystFEL geometry by first data coordinate
+
+        Index panels by which part of the data they refer to, rather than
+        relying on names like p0a0.
+        """
+        res = {}
+        for pname, info in panels.items():
+            dims = info['dim_structure']
+            ix_dims = [i for i in dims if isinstance(i, int)]
+            if len(ix_dims) > 1:
+                raise ValueError(f"Too many index dimensions for {pname}: {dims}")
+
+            min_ss = info['min_ss']
+            if ix_dims:
+                # Geometry for 3D data, modules stacked along separate axis
+                modno = ix_dims[0]
+            else:
+                # Geometry for 2D data, modules concatenated along slow-scan axis
+                modno, min_ss = divmod(min_ss, cls.expected_data_shape[1])
+
+            res[(modno, min_ss, info['min_fs'])] = info
+
+        return res
+
     @classmethod
     def from_crystfel_geom(cls, filename):
         """Read a CrystFEL format (.geom) geometry file.
@@ -240,19 +266,26 @@ class DetectorGeometryBase:
         Returns a new geometry object.
         """
         geom_dict = load_crystfel_geometry(filename)
+        panels_by_data_coord = cls._cfel_panels_by_data_coord(geom_dict['panels'])
+        n_modules = cls.n_modules
+        if n_modules == 0:
+            # Detector type with varying number of modules (e.g. JUNGFRAU)
+            n_modules = max(c[0] for c in panels_by_data_coord) + 1
+
         modules = []
-        for p in range(cls.n_modules):
+        for p in range(n_modules):
             tiles = []
             modules.append(tiles)
             for a in range(cls.n_tiles_per_module):
-                d = geom_dict['panels']['p{}a{}'.format(p, a)]
+                ss_slice, fs_slice = cls._tile_slice(a)
+                d = panels_by_data_coord[p, ss_slice.start, fs_slice.start]
                 tiles.append(GeometryFragment.from_panel_dict(d))
         return cls(modules, filename=filename)
 
     def write_crystfel_geom(self, filename, *,
                             data_path='/entry_1/instrument_1/detector_1/data',
                             mask_path=None, dims=('frame', 'modno', 'ss', 'fs'),
-                            nquads=4, adu_per_ev=None, clen=None,
+                            nquads=None, adu_per_ev=None, clen=None,
                             photon_energy=None):
         """Write this geometry to a CrystFEL format (.geom) geometry file.
 
@@ -280,6 +313,8 @@ class DetectorGeometryBase:
         photon_energy : float
             Beam wave length in eV
         """
+        if nquads is None:
+            nquads = self.n_quads
         write_crystfel_geom(
             self, filename, data_path=data_path, mask_path=mask_path, dims=dims,
             nquads=nquads, adu_per_ev=adu_per_ev, clen=clen,
@@ -347,9 +382,10 @@ class DetectorGeometryBase:
         Parameters
         ----------
 
-        data : ndarray
+        data : ndarray or xarray.DataArray
           The last three dimensions should match the modules, then the
-          slow scan and fast scan pixel dimensions.
+          slow scan and fast scan pixel dimensions. If an xarray labelled array
+          is given, it must have a 'module' dimension.
         out : ndarray, optional
           An output array to assemble the image into. By default, a new
           array is allocated. Use :meth:`output_array_for_position_fast` to
@@ -389,9 +425,10 @@ class DetectorGeometryBase:
         Parameters
         ----------
 
-        data : ndarray
+        data : ndarray or xarray.DataArray
           The last three dimensions should match the modules, then the
-          slow scan and fast scan pixel dimensions.
+          slow scan and fast scan pixel dimensions. If an xarray labelled array
+          is given, it must have a 'module' dimension.
         out : ndarray, optional
           An output array to assemble the image into. By default, a new
           array is created at the minimum size to allow symmetric assembly.
@@ -625,7 +662,7 @@ class DetectorGeometryBase:
 
         .. seealso::
 
-           :doc:`agipd_geometry` demonstrates using this method.
+           :doc:`convert_coords` demonstrates using this method.
         """
         assert module_no.shape == slow_scan.shape == fast_scan.shape
 
@@ -744,6 +781,7 @@ class AGIPD_1MGeometry(DetectorGeometryBase):
     frag_ss_pixels = 64
     frag_fs_pixels = 128
     expected_data_shape = (16, 512, 128)
+    n_quads = 4
     n_modules = 16
     n_tiles_per_module = 8
 
@@ -869,6 +907,7 @@ class AGIPD_1MGeometry(DetectorGeometryBase):
         centre : ndarray
           (y, x) pixel location of the detector centre in this geometry.
         """
+        from scipy.ndimage import affine_transform
         assert data.shape == (16, 512, 128)
         size_yx, centre = self._get_dimensions()
         tmp = np.empty((16 * 8,) + size_yx, dtype=data.dtype)
@@ -1156,6 +1195,7 @@ class LPD_1MGeometry(DetectorGeometryBase):
     pixel_size = 5e-4  # 5e-4 metres == 0.5 mm
     frag_ss_pixels = 32
     frag_fs_pixels = 128
+    n_quads = 4
     n_modules = 16
     n_tiles_per_module = 16
     expected_data_shape = (16, 256, 256)
@@ -1535,6 +1575,7 @@ class DSSC_1MGeometry(DetectorGeometryBase):
     pixel_size = 236e-6
     frag_ss_pixels = 128
     frag_fs_pixels = 256
+    n_quads = 4
     n_modules = 16
     n_tiles_per_module = 2
     expected_data_shape = (16, 128, 512)
@@ -1842,6 +1883,8 @@ class JUNGFRAUGeometry(DetectorGeometryBase):
     pixel_size = 7.5e-5   # 7.5e-5 metres = 75 micrometer = 0.075 mm
     frag_ss_pixels = 256  # pixels along slow scan axis within tile
     frag_fs_pixels = 256  # pixels along fast scan axis within tile
+    expected_data_shape = (0, 512, 1024)  # num modules filled at instantiation
+    n_tiles_per_module = 8
 
     def __init__(self, modules, filename='No file'):
         super().__init__(modules, filename)
@@ -1861,7 +1904,7 @@ class JUNGFRAUGeometry(DetectorGeometryBase):
           for each offset to the global origin. Coordinates are in pixel units
           by default.
 
-          These offsets are positions for the bottom, beam-left corner of each
+          These offsets are positions for the bottom, beam-right corner of each
           module, regardless of its orientation.
 
         orientations: iterable of tuples
@@ -1971,15 +2014,6 @@ class JUNGFRAUGeometry(DetectorGeometryBase):
         fs_slice = slice(tile_fs_offset, tile_fs_offset + cls.frag_fs_pixels)
         return ss_slice, fs_slice
 
-    @classmethod
-    def from_crystfel_geom(cls, filename):
-
-        raise NotImplementedError
-
-    def write_crystfel_geom(self, file_name):
-
-        raise NotImplementedError
-
 
 class PNCCDGeometry(DetectorGeometryBase):
     """Detector layout for pnCCD
@@ -1991,7 +2025,7 @@ class PNCCDGeometry(DetectorGeometryBase):
     out and written to file as a single image, with the modules split
     along the slow-scan dimension y. The public methods of this type
     support both the combined image array as well as separated module
-    with expected_data_shape.
+    with :attr:`expected_data_shape`.
     """
 
     detector_type_name = 'PNCCD1MP'
